#include "mesh.h"
#include <QFile>
#include <QStringList>
#include <QList>
#include "utils.h"
#include <unordered_map>
#include <fstream>
#include <QGLWidget>

namespace std
{
	template<>
	struct hash<Vec3> : public hash<float>
	{
		inline size_t operator()(const Vec3 &v) const
		{
			return (hash<float>::operator ()(v.x()) * 31U + hash<float>::operator ()(v.y())) * 31U + hash<float>::operator ()(v.z());
		}
	};
}

#define PLY_NONE        0x0
#define PLY_VERTEX      0x1
#define PLY_FACE        0x2

#define FORMAT_ASCII                0x0
#define FORMAT_BINARY_BIG_ENDIAN    0x1
#define FORMAT_BINARY_LITTLE_ENDIAN 0x2

using namespace std;

namespace
{
	union INT
	{
		int i;
		u_int8_t b[4];
	};

	union FLOAT
	{
		float f;
		u_int8_t b[4];
	};

	union SHORT
	{
		short s;
		u_int8_t b[2];
	};

	inline int swap_int(int a, u_int8_t file_format)
	{
		if (file_format == FORMAT_BINARY_LITTLE_ENDIAN)    return a;
		INT A, B;
		A.i = a;
		B.b[0] = A.b[3];
		B.b[1] = A.b[2];
		B.b[2] = A.b[1];
		B.b[3] = A.b[0];
		return B.i;
	}

	inline short swap_short(short a, u_int8_t file_format)
	{
		if (file_format == FORMAT_BINARY_LITTLE_ENDIAN)    return a;
		SHORT A, B;
		A.s = a;
		B.b[0] = A.b[1];
		B.b[1] = A.b[0];
		return B.s;
	}

	inline float swap_float(float a, u_int8_t file_format)
	{
		if (file_format == FORMAT_BINARY_LITTLE_ENDIAN)    return a;
		FLOAT A, B;
		A.f = a;
		B.b[0] = A.b[3];
		B.b[1] = A.b[2];
		B.b[2] = A.b[1];
		B.b[3] = A.b[0];
		return B.f;
	}
}

void Mesh::loadPLY(const QString &filename)
{
	QFile file(filename);

	file.open(QIODevice::ReadOnly | QIODevice::Text);
	if (!file.isOpen())
		return;

	if (file.readLine().left(3) != "ply")
		return;

	primitive_type = GL_TRIANGLES;
	QList<const char*> params;
	QList<int> vertex_params;
	QList<int> vertex_types;
	int nb_face_params = 0;
	QList<int> face_types;

	enum { T_FLOAT, T_INT, T_UCHAR, T_LIST, T_UNKNOWN };
	enum { P_X, P_Y, P_Z, P_R, P_G, P_B, P_UNKNOWN };

	bool header = true;
	u_int8_t element_type = PLY_NONE;
	u_int8_t file_format = FORMAT_ASCII;

	u_int32_t current_vertex = 0;
	u_int32_t current_face = 0;

	u_int32_t nb_faces(0);
	char line[1024];

	do
	{
		if (header || file_format == FORMAT_ASCII)
		{
			const quint64 len = file.readLine(line, 1024);
			params.clear();
			bool bPrevEmpty = true;
			for(size_t i = 0 ; i < len ; ++i)
			{
				if (bPrevEmpty && !isspace(line[i]))
				{
					params.push_back(line + i);
					bPrevEmpty = false;
				}
				else
				{
					bPrevEmpty = isspace(line[i]);
					if (bPrevEmpty)
						line[i] = 0;
				}
			}

			if (params.isEmpty()) continue;
		}

		if (header)
		{
			if (!strcmp(params[0], "end_header"))
			{
				header = false;
				if (file_format == FORMAT_BINARY_BIG_ENDIAN || file_format == FORMAT_BINARY_LITTLE_ENDIAN)
					file.setTextModeEnabled(false);
			}
			else if (!strcmp(params[0], "format") && params.size() > 1)
			{
				if (!strcmp(params[1], "binary_big_endian"))
					file_format = FORMAT_BINARY_BIG_ENDIAN;
				else if (!strcmp(params[1], "binary_little_endian"))
					file_format = FORMAT_BINARY_LITTLE_ENDIAN;
				else
					file_format = FORMAT_ASCII;
			}
			else if (!strcmp(params[0], "element") && params.size() > 1)
			{
				if (!strcmp(params[1], "vertex") && params.size() > 2)
				{
					vertex.resize(atoi(params[2]));
					element_type = PLY_VERTEX;
				}
				else if (!strcmp(params[1], "face") && params.size() > 2)
				{
					nb_faces = atoi(params[2]);
					index.reserve(nb_faces * 3);
					element_type = PLY_FACE;
				}
			}
			else if (!strcmp(params[0], "property") && params.size() > 1)
			{
				switch(element_type)
				{
				case PLY_VERTEX:
					if (!strcmp(params[2], "x"))		vertex_params.push_back( P_X );
					else if (!strcmp(params[2], "y"))	vertex_params.push_back( P_Y );
					else if (!strcmp(params[2], "z"))	vertex_params.push_back( P_Z );
					else if (!strcmp(params[2], "red"))	vertex_params.push_back( P_R );
					else if (!strcmp(params[2], "green"))	vertex_params.push_back( P_G );
					else if (!strcmp(params[2], "blue"))	vertex_params.push_back( P_B );
					else						vertex_params.push_back(P_UNKNOWN);
					if (!strcmp(params[1], "float"))	vertex_types.push_back( T_FLOAT );
					else if (!strcmp(params[1], "int"))	vertex_types.push_back( T_INT );
					else if (!strcmp(params[1], "uchar"))	vertex_types.push_back( T_UCHAR );
					else if (!strcmp(params[1], "list"))	vertex_types.push_back( T_LIST );
					else							vertex_types.push_back( T_UNKNOWN );
					break;
				case PLY_FACE:
					if (!strcmp(params[1], "list"))
					{
						if (!strcmp(params[1], "float"))	face_types.push_back( T_FLOAT );
						else if (!strcmp(params[1], "int"))	face_types.push_back( T_INT );
						else if (!strcmp(params[1], "uchar"))	face_types.push_back( T_UCHAR );
						else if (!strcmp(params[1], "list"))	face_types.push_back( T_LIST );
						else							face_types.push_back( T_UNKNOWN );

						++nb_face_params;
						++nb_face_params;
						++nb_face_params;
						if (!strcmp(params[2], "float"))	face_types.push_back( T_FLOAT );
						else if (!strcmp(params[2], "int"))	face_types.push_back( T_INT );
						else if (!strcmp(params[2], "uchar"))	face_types.push_back( T_UCHAR );
						else if (!strcmp(params[2], "list"))	face_types.push_back( T_LIST );
						else							face_types.push_back( T_UNKNOWN );
						if (!strcmp(params[3], "float"))	face_types.push_back( T_FLOAT );
						else if (!strcmp(params[3], "int"))	face_types.push_back( T_INT );
						else if (!strcmp(params[3], "uchar"))	face_types.push_back( T_UCHAR );
						else if (!strcmp(params[3], "list"))	face_types.push_back( T_LIST );
						else							face_types.push_back( T_UNKNOWN );
					}
					else
					{
						++nb_face_params;
						if (!strcmp(params[1], "float"))	face_types.push_back( T_FLOAT );
						else if (!strcmp(params[1], "int"))	face_types.push_back( T_INT );
						else if (!strcmp(params[1], "uchar"))	face_types.push_back( T_UCHAR );
						else if (!strcmp(params[1], "list"))	face_types.push_back( T_LIST );
						else							face_types.push_back( T_UNKNOWN );
					}
					break;
				};
			}
		}
		else
		{
			switch(file_format)
			{
			case FORMAT_ASCII:
				if (current_vertex < vertex.size() && params.size() == vertex_params.size())
				{
					float x = 0.f, y = 0.f, z = 0.f;
					for(int i = 0 ; i < vertex_params.size() ; i++)
						switch(vertex_params[i])
						{
						case P_X:	x = parse_float(params[i]);	break;
						case P_Y:	y = parse_float(params[i]);	break;
						case P_Z:	z = parse_float(params[i]);	break;
                        case P_R:	break;
                        case P_G:	break;
                        case P_B:	break;
						}
					vertex[current_vertex++] = Vec3(x, y, z);
				}
				else if (current_face < nb_faces)
				{
					u_int32_t nb = atoi(params[0]);
					for(u_int32_t i = 2 ; i < min( nb, (u_int32_t)params.size() - 1 ) ; i++)
					{
						index.push_back(parse_int(params[1]));
						index.push_back(parse_int(params[i]));
						index.push_back(parse_int(params[i+1]));
					}
					current_face++;
				}
				break;
			case FORMAT_BINARY_BIG_ENDIAN:
			case FORMAT_BINARY_LITTLE_ENDIAN:
				if (current_vertex < vertex.size())
				{
					for(current_vertex = 0 ; current_vertex < vertex.size() ; ++current_vertex)
					{
						float x = 0.f, y = 0.f, z = 0.f;
						for(int i = 0 ; i < vertex_params.size() ; i++)
						{
							float v;
							switch(vertex_types[i])
							{
							case T_FLOAT:
								{
									file.read((char*)&v,sizeof(v));
									v = swap_float(v, file_format);
								}
								break;
							case T_UCHAR:
								{
									u_int8_t vub;
									file.read((char*)&vub, sizeof(vub));
									v = vub;
								}
								break;
							case T_INT:
								{
									int vi;
									file.read((char*)&vi, sizeof(vi));
									v = swap_int(vi, file_format);
								}
								break;
							}

							switch(vertex_params[i])
							{
							case P_X:	x = v;	break;
							case P_Y:	y = v;	break;
							case P_Z:	z = v;	break;
                            case P_R:	break;
                            case P_G:	break;
                            case P_B:	break;
							}
						}
						vertex[current_vertex] = Vec3(x, y, z);
					}
				}
				else if (current_face < nb_faces)
				{
					for(current_face = 0 ; current_face < nb_faces ; ++current_face)
					{
						for(int i = 0 ; i < nb_face_params ; i++)
						{
							float v;
							const bool list_mode = face_types[i] == T_LIST;
							if (list_mode)
								++i;
							switch(face_types[i])
							{
							case T_FLOAT:
								{
									file.read((char*)&v, sizeof(float));
									v = swap_float(v, file_format);
								}
								break;
							case T_UCHAR:
								{
									u_int8_t vub;
									file.read((char*)&vub, sizeof(vub));
									v = vub;
								}
								break;
							case T_INT:
								{
									int vi;
									file.read((char*)&vi, sizeof(vi));
									v = swap_int(vi, file_format);
								}
								break;
							}

							if (list_mode)
							{
								++i;
								int nb = (int)v;
								u_int32_t a, b, c;
								for(int e = 0 ; e < nb ; ++e)
								{
									if (face_types[i] == T_INT)
									{
										int vi;
										file.read((char*)&vi, sizeof(vi));
										c = swap_int(vi, file_format);
										if (e == 0)
											a = c;
									}
									if (e >= 2)
									{
										index.push_back(a);
										index.push_back(b);
										index.push_back(c);
									}
									b = c;
								}
							}
						}
					}
				}
				else
					file.close();
				break;
			};
		}

	}while(file.isOpen() && !file.atEnd());

	file.close();

	recompute_normals();
}
