#include "simpletextedit.h"
#include <QPaintEvent>
#include <QPainter>
#include <QWheelEvent>

SimpleTextEdit::SimpleTextEdit(QWidget *parent) :
    QWidget(parent)
{
    scroll_position = 0;
    max_scroll = 0;
	max_size = size_t(-1);
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    font_height = 0;
    font_width = 0;
}

void SimpleTextEdit::setPlainText(const QByteArray &text)
{
    data = text.split('\n');
	const int h = data.size() * font_height;
    max_scroll = h;
    int w = 0;
    foreach(const QByteArray &line, data)
        w = qMax(w, line.size());
	w *= font_width;

    scroll_position = 0;

    update();

    emit maxScrollChanged(qMax(0, max_scroll - height()));
}

QByteArray SimpleTextEdit::toPlainText() const
{
    QByteArray ret;
    ret.reserve(data.size() * 20);
    for(int i = 0 ; i < data.size() ; ++i)
    {
        if (i > 0)
            ret.append('\n');
        ret.append(data[i]);
    }
    return ret;
}

void SimpleTextEdit::paintEvent(QPaintEvent *e)
{
    QPainter painter(this);
    painter.setPen(palette().text().color());
    if (font_width == 0)
        font_width = QPainter().fontMetrics().averageCharWidth();
    if (font_height == 0)
    {
        font_height = painter.fontMetrics().height();
        emit maxScrollChanged(getMaxScroll());
    }
    const int h = font_height;
    foreach(const QRect &r, e->region().rects())
    {
        painter.fillRect(r, palette().base());
        const int first_line = (r.top() + scroll_position) / h;
        const int end_line = (r.bottom() + scroll_position + h - 1) / h;
        for(int i = first_line ; i < end_line ; ++i)
        {
            if (i >= data.size())
                break;
            painter.drawText(0, (i + 1) * h - scroll_position, data[i]);
        }
    }
}

int SimpleTextEdit::getScrollPosition() const
{
    return scroll_position;
}

void SimpleTextEdit::setScrollPosition(const int pos)
{
    if (scroll_position != pos)
    {
        scroll_position = pos;
        update();
        emit scrollPositionChanged(scroll_position);
    }
}

int SimpleTextEdit::getMaxScroll() const
{
    return qMax(0, max_scroll - height());
}

void SimpleTextEdit::wheelEvent(QWheelEvent *e)
{
    const int last_pos = scroll_position;
    scroll_position = qMax(0, qMin(max_scroll - height(), scroll_position - e->delta()));
    if (last_pos != scroll_position)
    {
        update();
        emit scrollPositionChanged(scroll_position);
    }
}

void SimpleTextEdit::resizeEvent(QResizeEvent *e)
{
    QWidget::resizeEvent(e);
    emit maxScrollChanged(qMax(0, scroll_position - height()));
}

void SimpleTextEdit::append(const QByteArray &text)
{
	const int h = font_height;

    data.push_back(text);
	if (data.size() >= max_size)
	{
		data.pop_front();
		update();
		return;
	}
    const bool b_at_end = scroll_position == getMaxScroll();
    max_scroll += h;
    emit maxScrollChanged(getMaxScroll());
    if (b_at_end)
        scroll_position += h;

    emit maxScrollChanged(getMaxScroll());
    update();
    if (b_at_end)
        emit scrollPositionChanged(scroll_position);
}

void SimpleTextEdit::popFront()
{
    if (data.isEmpty())
        return;

	const int h = font_height;

    data.pop_front();
    scroll_position = qMax(0, scroll_position - h);
    max_scroll = qMax(0, max_scroll - h);

    emit maxScrollChanged(getMaxScroll());
    emit scrollPositionChanged(scroll_position);

    update();
}

size_t SimpleTextEdit::getSize() const
{
	return data.size();
}

void SimpleTextEdit::setMaxSize(size_t max_size)
{
	this->max_size = max_size;
	if (data.size() > max_size)
	{
		data.erase(data.begin(), data.begin() + (data.size() - max_size));

		const int h = font_height;

		scroll_position = qMax(0, scroll_position - h);
		max_scroll = qMax(0, max_scroll - h);

		emit maxScrollChanged(getMaxScroll());
		emit scrollPositionChanged(scroll_position);

		update();
	}
}
